#include <sys/param.h>
#include <sys/stdint.h>
#include <sys/types.h>
#include <sys/queue.h>

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <libutil.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <stdarg.h>
#include <unistd.h>

#ifndef TAG
#define TAG "polldev"
#endif

char	pidfile[PATH_MAX] = "/var/run/" TAG ".pid";
char   *ident = TAG, *basedir = NULL;
int	bd_namlen;
int	facility = LOG_DAEMON;
int	delay = 500;
int	open_flags = O_WRONLY;
int	pe = 1;
int	enable_scan = 1;

static void
logmsg(int priority, const char *fmt, ...)
{
  static int need_open = 1;
  va_list ap, apb;

  if (need_open) {
    openlog(ident, LOG_PID, facility);
    need_open = 0;
  }

  va_start(ap, fmt);

  if (pe) {
    va_copy(apb, ap);
    fprintf(stderr, TAG ": ");
    vfprintf(stderr, fmt, apb);
    va_end(apb);
    fprintf(stderr, "\n");
  }

  vsyslog(priority, fmt, ap);
  va_end(ap);
}

static void
usage(void)
{
  logmsg(LOG_ERR, "usage: " TAG " [-i interval] [-m] [-n] [-p pidfile] device [device ...]");
  exit(EX_USAGE);
}

typedef struct {
  char *name;
  int namlen;
  int take_round;
} dev;

struct devlist {
  TAILQ_ENTRY(devlist) next;
  dev buffer;
};

TAILQ_HEAD(DLH, devlist);
static struct DLH head;

int
check_open(const char *name)
{
  int fd;

  if ((fd = open(name, open_flags)) < 0)
  switch (errno) {
    case EWOULDBLOCK:
    case ENXIO:
	break;
    default:
      return 1;
  }

  if (fd >=0)
    close(fd);
  return 0;
}

int
add_device(char *name)
{
  struct devlist *p;

  if(check_open(name)) {
      logmsg(LOG_ERR, "%s excluded from the list: %s", name, strerror(errno));
      return 1;
  }

  if ((p = malloc(sizeof(*p))) == NULL) {
    logmsg(LOG_ERR, "out of memory allocating %u bytes", (unsigned)sizeof(*p));
    return 1;
  }
  p->buffer.name = name;
/*p->buffer.namlen = strlen(name); */
  TAILQ_INSERT_TAIL(&head, p, next);
  return 0;
}

void shutdown_daemon(int s)
{
  unlink(pidfile);
  logmsg(LOG_NOTICE, "terminated with signal %d", s);
  exit(0);
}

int main(int argc, char *argv[])
{
  struct pidfh *pfh;
  struct devlist *p, *temp;
  struct dirent *de;
  DIR *d;
  char *s;
  pid_t otherpid;
  int i;

  while ((i = getopt(argc, argv, "hi:mnp:")) != -1)
  switch (i) {
    case 'i':
      delay = atoi(optarg);
      if (delay < 1)
        delay = 1;
      break;
    case 'm':
      enable_scan = 0;
      break;
    case 'n':
      open_flags |= O_NONBLOCK;
      break;
    case 'p':
      strlcpy(pidfile, optarg, sizeof(pidfile));
      break;
    case 'h':
    default: usage();
  }
  argc -= optind;
  argv += optind;

  if (argc == 0)
    usage();

  if ((pfh = pidfile_open(pidfile, 0666, &otherpid)) == NULL) {
    if (errno == EEXIST) {
      logmsg(LOG_ERR, "already running, pid: %jd", (intmax_t)otherpid);
      exit(EX_UNAVAILABLE);
    }
    else
      logmsg(LOG_ERR, "cannot create pidfile: %s", strerror(errno));
  }

  TAILQ_INIT(&head);
  for (i=0; i < argc; i++) {
    add_device(argv[i]);
  }

  if (TAILQ_EMPTY(&head)) {
    logmsg(LOG_ERR, "no existing device given, exiting");
    exit(EX_DATAERR);
  }

  if (enable_scan && (s = strdup(TAILQ_FIRST(&head)->buffer.name))) {
    basedir = dirname(s);
    bd_namlen = strlen(basedir) + 1;
    free(s);
    TAILQ_FOREACH_SAFE(p, &head, next, temp) {
	p->buffer.namlen = strlen(p->buffer.name) - bd_namlen;
      }
  }

  if (daemon(0, 0) == -1) {
    logmsg(LOG_ERR, "cannot daemonize, exiting");
    pidfile_remove(pfh);
    exit(EXIT_FAILURE);
  }

  pe = 0;
  pidfile_write(pfh);
  pidfile_close(pfh);
  logmsg(LOG_INFO, "created pidfile %s", pidfile);
  signal(SIGINT, &shutdown_daemon);
  signal(SIGTERM, &shutdown_daemon);

  delay *= 1000;

  /* main loop */
  while(1) {
    if(basedir) {
      TAILQ_FOREACH_SAFE(p, &head, next, temp) {
        p->buffer.take_round = 1;
      }

      if ((d = opendir(basedir))) {
        while((de = readdir(d))) {
          TAILQ_FOREACH_SAFE(p, &head, next, temp) {
            if (p->buffer.take_round && de->d_namlen > p->buffer.namlen &&
                !strncmp(de->d_name, p->buffer.name + bd_namlen, p->buffer.namlen)) {
    	      p->buffer.take_round = 0;
    	      logmsg(LOG_DEBUG, "%s skipped round due to found %s", p->buffer.name, de->d_name);
            }
          }
        }
        closedir(d);
      }
    }

    TAILQ_FOREACH_SAFE(p, &head, next, temp) {
      if (((basedir && p->buffer.take_round) || !basedir) &&
           check_open(p->buffer.name)) {
        logmsg(LOG_NOTICE, "removed from checklist %s: %m",
    	    p->buffer.name);
    	TAILQ_REMOVE(&head, p, next);
    	free(p);
      }
    }
    if (TAILQ_EMPTY(&head)) {
      logmsg(LOG_NOTICE, "checklist is empty, exiting");
      unlink(pidfile);
      exit(0);
    }
    usleep(delay);
  }
}
